"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DownloadManager = exports.EVENTS = exports.DOWNLOAD_STATUS = void 0;
const eventemitter3_1 = __importDefault(require("eventemitter3"));
const v4_1 = __importDefault(require("uuid/v4"));
const Playlist_1 = require("./Playlist");
Object.defineProperty(exports, "DOWNLOAD_STATUS", { enumerable: true, get: function () { return Playlist_1.DOWNLOAD_STATUS; } });
exports.EVENTS = {
    CHANGE: "CHANGE",
    PROGRESS: "PROGRESS",
    ACTION_COMPLETE: "ACTION_COMPLETE",
};
let totalBytesAvailable = Infinity;
let _actionId = 0;
function actionId() {
    return `${_actionId++}`;
}
class DownloadManager {
    constructor(storage, fileSystem, options = {}) {
        this.parallelDownloadCount = 3;
        this.didSetPlaylist = false;
        this._playlistActionQueue = [];
        this._handleProgress = (uri, downloadProgress) => {
            var _a;
            if (!downloadProgress)
                return;
            if (downloadProgress.totalBytesExpectedToWrite >=
                totalBytesAvailable - 10 * 1000 * 1000) {
                console.log("DownloadManager: OUT OF MEMORY");
                this.pause();
                (_a = this._onFatalError) === null || _a === void 0 ? void 0 : _a.call(this, new Error("Dart Disco ran out of memory!"));
                return;
            }
            const itemMetadata = this._playlist.getMetadata(uri);
            if (!itemMetadata)
                return;
            itemMetadata.totalBytesWritten = downloadProgress.totalBytesWritten;
            itemMetadata.totalBytesExpectedToWrite =
                downloadProgress.totalBytesExpectedToWrite;
            itemMetadata.progress =
                downloadProgress.totalBytesWritten /
                    downloadProgress.totalBytesExpectedToWrite;
            this._emitter.emit(exports.EVENTS.PROGRESS);
        };
        this._emitter = new eventemitter3_1.default();
        this.storage = storage;
        this.fileSystem = fileSystem;
        this._platform = options.platform;
        this._onFatalError = options.onFatalError;
        this.storageKey = options.storageKey
            ? options.storageKey
            : options.downloadDirectory
                ? `${Playlist_1.DEFAULT_STORAGE_KEY}-${options.downloadDirectory}`
                : Playlist_1.DEFAULT_STORAGE_KEY;
        if (options.parallelDownloadCount) {
            this.parallelDownloadCount = options.parallelDownloadCount;
        }
        this._playlist = new Playlist_1.Playlist(undefined, storage, fileSystem, {
            storageKey: this.storageKey,
            platform: options.platform,
            downloadDirectory: options.downloadDirectory,
        });
        this._updateTotalBytesAvailable();
    }
    get didRestore() {
        return this._playlist.didRestore;
    }
    get ready() {
        return this._playlist.ready;
    }
    /**
     * Add a new playlist action to the queue
     */
    _addPlaylistAction(action) {
        switch (action.type) {
            case "setPlaylist":
                // this._playlistActionQueue.filter(
                //   (action) => action.type !== "setPlaylist"
                // );
                break;
            case "download":
                break;
            case "pause":
                if (this._playlistAction && this._playlistAction.type === "download") {
                    this._playlistAction = undefined;
                }
                // this._playlistActionQueue.filter(
                //   (action) => action.type !== "download"
                // );
                break;
            case "delete":
                if (this._playlistAction && this._playlistAction.type === "download") {
                    this._playlistAction = undefined;
                }
                this._playlistActionQueue.length = 0;
                break;
        }
        this._playlistActionQueue.push(action);
    }
    /**
     * Starting the queue if it hasn't already started
     */
    async _runPlaylistActionQueue() {
        if (this._playlistAction || this._playlistActionQueue.length === 0)
            return;
        await this._runNextPlaylistAction();
        this._runPlaylistActionQueue();
    }
    async _runNextPlaylistAction() {
        if (this._playlistAction) {
            throw new Error("Can't start another action while one is in progress");
        }
        const action = this._playlistActionQueue.shift();
        if (!action)
            return;
        this._playlistAction = action;
        switch (action.type) {
            case "setPlaylist":
                await this._setPlaylist(action.items);
                break;
            case "download":
                await this._start();
                break;
            case "pause":
                await this._pause();
                break;
            case "delete":
                await this._delete();
                break;
        }
        this._playlistAction = undefined;
        this._emitter.emit(exports.EVENTS.ACTION_COMPLETE, action);
    }
    async _setPlaylist(items) {
        if (this._platform === "web")
            return;
        this._updateTotalBytesAvailable();
        await this.didRestore;
        this._playlist.setItems(items);
        this.didSetPlaylist = true;
        // We don't wait for deleting files to finish, since we'll continue regardless of whether
        // it succeeds or not
        this._playlist.deleteUnusedFiles();
        this._emitter.emit(exports.EVENTS.CHANGE);
        await this._playlist.save();
        return;
    }
    async _start() {
        if (this._platform === "web")
            return;
        let parallelDownloads = {};
        while (true) {
            const nextItem = this._getNextItemToDownload();
            // If there are no more items, and no current downloads, return
            if (!nextItem && Object.keys(parallelDownloads).length === 0)
                return;
            // Add the next item to the parallelDownloads
            if (nextItem &&
                Object.keys(parallelDownloads).length < this.parallelDownloadCount) {
                const downloadItem = nextItem;
                parallelDownloads[downloadItem] = this._startFile(downloadItem).then(() => downloadItem);
            }
            else {
                // Download all items
                const finishedItem = await Promise.race(Object.values(parallelDownloads));
                delete parallelDownloads[finishedItem];
                if (!this._playlistAction || this._playlistAction.type !== "download") {
                    break;
                }
            }
        }
    }
    async _pause() {
        if (this._platform === "web")
            return;
        const promises = this._playlist.items.map((uri) => this.pauseFile(uri));
        try {
            await Promise.all(promises);
        }
        catch (e) {
            console.log("DownloadManager: Failed to pause downloads", e);
        }
    }
    async _delete() {
        if (this._platform === "web")
            return;
        const promises = this._playlist.items.map((uri) => this.deleteFile(uri));
        try {
            await Promise.all(promises);
        }
        catch (e) {
            console.log("DownloadManager: Failed to delete downloads", e);
        }
    }
    _awaitAction(action) {
        return new Promise((resolve) => {
            this._emitter.on(exports.EVENTS.ACTION_COMPLETE, (completed) => {
                if (completed.id === action.id) {
                    resolve(completed);
                }
            });
        });
    }
    // PUBLIC
    setPlaylist(items) {
        const action = {
            type: "setPlaylist",
            items,
            id: actionId(),
        };
        this._addPlaylistAction(action);
        this._runPlaylistActionQueue();
        return this._awaitAction(action);
    }
    start() {
        const action = {
            type: "download",
            id: actionId(),
        };
        this._addPlaylistAction(action);
        this._runPlaylistActionQueue();
        return this._awaitAction(action);
    }
    pause() {
        const action = {
            type: "pause",
            id: actionId(),
        };
        this._addPlaylistAction(action);
        this._runPlaylistActionQueue();
        return this._awaitAction(action);
    }
    delete() {
        const action = {
            type: "delete",
            id: actionId(),
        };
        this._addPlaylistAction(action);
        this._runPlaylistActionQueue();
        return this._awaitAction(action);
    }
    on(event, fn, context) {
        this._emitter.on(event, fn, context);
        return {
            remove: () => {
                this.removeListener(event, fn, context);
            },
        };
    }
    removeListener(event, fn, context) {
        return this._emitter.removeListener(event, fn, context);
    }
    /**
     * Get the status of the initial playlist restored from storage
     */
    getRestoredStatus() {
        return this._playlist.getStatus();
    }
    /**
     * Get the status of the current playlist
     */
    getStatus() {
        return {
            ...this._playlist.getStatus(),
            totalBytesAvailable,
        };
    }
    getSummary() {
        const status = this.getStatus();
        return `DownloadManager | Files: ${status.files.length} | Progress: ${status.progress} | Playlist: ${status.status}`;
    }
    async startFile(item) {
        await this.didRestore;
        return this._startFile(item);
    }
    // Immediately change the status of the item from NOT_STARTED, removing
    // it from the queue of subsequence items to download
    async _startFile(item) {
        if (this._platform === "web")
            return;
        this._updateTotalBytesAvailable();
        if (!this._playlist.hasItem(item)) {
            throw new Error(`DownloadManager: Failed to start, uri '${item}' not in playlist`);
        }
        const itemMetadata = this._playlist.getMetadata(item);
        switch (itemMetadata.downloadStatus) {
            case Playlist_1.DOWNLOAD_STATUS.IN_PROGRESS:
                itemMetadata.downloadStatus = Playlist_1.DOWNLOAD_STATUS.PAUSED;
                break;
            case Playlist_1.DOWNLOAD_STATUS.COMPLETED:
                return;
            case Playlist_1.DOWNLOAD_STATUS.FAILED:
                console.log("TODO: Retry download");
                return;
        }
        const resumableDownload = this._playlist.getResumableDownload(item) ||
            this._createResumableDownloadForItem(item);
        this._playlist.setResumableDownload(item, resumableDownload);
        // Start or resume the download
        let promise = undefined;
        switch (itemMetadata.downloadStatus) {
            case Playlist_1.DOWNLOAD_STATUS.NOT_STARTED:
                promise = resumableDownload.downloadAsync();
                break;
            case Playlist_1.DOWNLOAD_STATUS.PAUSED:
                promise = resumableDownload.resumeAsync();
                break;
            default:
                throw new Error(`Failed to start download. Invalid status for ${item}: ${itemMetadata.downloadStatus}`);
        }
        this._playlist.setDownloadStatus(item, Playlist_1.DOWNLOAD_STATUS.IN_PROGRESS);
        promise = promise.then((download) => {
            if (!download) {
                console.log("DownloadManager: Stopped/paused", item, this._playlist.getDownloadStatus(item));
                return undefined;
            }
            const { uri: fileUri, status } = download;
            const itemMetadata = this._playlist.getMetadata(item);
            itemMetadata.downloadStatus =
                status >= 200 && status < 300
                    ? Playlist_1.DOWNLOAD_STATUS.COMPLETED
                    : Playlist_1.DOWNLOAD_STATUS.FAILED;
            itemMetadata.fileUri = fileUri;
            delete itemMetadata.savedDownload;
            this._playlist.setResumableDownload(item, undefined);
            this._playlist.save();
            this._emitter.emit(exports.EVENTS.CHANGE);
        });
        try {
            await promise;
        }
        catch (e) {
            console.log(`DownloadManager: Failed to start downloading ${item}`, e);
        }
    }
    async pauseFile(item) {
        if (this._platform === "web")
            return;
        await this.didRestore;
        if (!this._playlist.hasItem(item)) {
            throw new Error(`DownloadManager: Failed to pause, uri '${item}' not in playlist`);
        }
        const itemMetadata = this._playlist.getMetadata(item);
        const resumableDownload = this._playlist.getResumableDownload(item);
        if (itemMetadata.downloadStatus === Playlist_1.DOWNLOAD_STATUS.COMPLETED ||
            itemMetadata.downloadStatus === Playlist_1.DOWNLOAD_STATUS.FAILED ||
            itemMetadata.downloadStatus === Playlist_1.DOWNLOAD_STATUS.PAUSED) {
            return;
        }
        if (resumableDownload) {
            this._playlist.setDownloadStatus(item, Playlist_1.DOWNLOAD_STATUS.PAUSED);
            try {
                await resumableDownload.pauseAsync();
            }
            catch (e) {
                console.log("DownloadManager: Failed to pause download", e);
            }
        }
        else {
            this._playlist.setDownloadStatus(item, Playlist_1.DOWNLOAD_STATUS.NOT_STARTED);
        }
        this._playlist.save();
        this._emitter.emit(exports.EVENTS.CHANGE);
    }
    async deleteFile(item) {
        if (this._platform === "web")
            return;
        await this.didRestore;
        if (!this._playlist.hasItem(item))
            return;
        const itemMetadata = this._playlist.getMetadata(item);
        if (itemMetadata.downloadStatus === Playlist_1.DOWNLOAD_STATUS.IN_PROGRESS) {
            await this.pauseFile(item);
        }
        if (itemMetadata.fileUri) {
            // Attempt to delete, even if the file does't exist
            this.fileSystem.deleteAsync(itemMetadata.fileUri, { idempotent: true });
            delete itemMetadata.fileUri;
        }
        this._playlist.setResumableDownload(item, undefined);
        this._playlist.setDownloadStatus(item, Playlist_1.DOWNLOAD_STATUS.NOT_STARTED);
        this._playlist.save();
        this._emitter.emit(exports.EVENTS.CHANGE);
        this._updateTotalBytesAvailable();
    }
    // BATCH ACTIONS
    _getNextItemToDownload() {
        const notCompletedFiles = [
            ...this._playlist.getItemsWithDownloadStatus(Playlist_1.DOWNLOAD_STATUS.NOT_STARTED),
            ...this._playlist.getItemsWithDownloadStatus(Playlist_1.DOWNLOAD_STATUS.PAUSED),
        ];
        return notCompletedFiles[0];
    }
    // PRIVATE
    _createResumableDownloadForItem(item) {
        const itemMetadata = this._playlist.getMetadata(item);
        if (itemMetadata.savedDownload) {
            const { url, fileUri, options, resumeData } = itemMetadata.savedDownload;
            if (url !== item) {
                throw new Error("Bad file downloader metadata");
            }
            return this.fileSystem.createDownloadResumable(url, fileUri, options, this._handleProgress.bind(this, item), resumeData);
        }
        else {
            const destination = `${this.fileSystem.documentDirectory}${this._playlist._downloadDirectory}/${v4_1.default()}.mp4`;
            return this.fileSystem.createDownloadResumable(item, destination, {}, this._handleProgress.bind(this, item));
        }
    }
    async _updateTotalBytesAvailable() {
        totalBytesAvailable = await this.fileSystem.getFreeDiskStorageAsync();
    }
}
exports.DownloadManager = DownloadManager;
