"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Playlist = exports.DEFAULT_DOWNLOAD_DIRECTORY = exports.DEFAULT_STORAGE_KEY = exports.getDocumentPaths = exports.DOWNLOAD_STATUS = void 0;
var DOWNLOAD_STATUS;
(function (DOWNLOAD_STATUS) {
    DOWNLOAD_STATUS["NOT_STARTED"] = "NOT_STARTED";
    DOWNLOAD_STATUS["IN_PROGRESS"] = "IN_PROGRESS";
    DOWNLOAD_STATUS["PAUSED"] = "PAUSED";
    DOWNLOAD_STATUS["COMPLETED"] = "COMPLETED";
    DOWNLOAD_STATUS["FAILED"] = "FAILED";
})(DOWNLOAD_STATUS = exports.DOWNLOAD_STATUS || (exports.DOWNLOAD_STATUS = {}));
async function getDocumentPaths(fileSystem, directory) {
    const documentDirectory = `${fileSystem.documentDirectory}${directory}/`;
    const files = await fileSystem.readDirectoryAsync(documentDirectory);
    return files.map((file) => documentDirectory + file);
}
exports.getDocumentPaths = getDocumentPaths;
function createDownloadItem(uri, metadata) {
    const { downloadStatus = DOWNLOAD_STATUS.NOT_STARTED, progress = 0, fileUri, totalBytesWritten = 0, totalBytesExpectedToWrite = 0, } = metadata || {};
    return {
        downloadStatus,
        progress,
        sourceUri: uri,
        fileUri,
        totalBytesWritten,
        totalBytesExpectedToWrite,
    };
}
function countFiles(items, status) {
    return items.reduce((result, item) => {
        return item.downloadStatus === status ? result + 1 : result;
    }, 0);
}
exports.DEFAULT_STORAGE_KEY = "downloadManagerMetadata";
exports.DEFAULT_DOWNLOAD_DIRECTORY = "react-native-file-downloader";
class Playlist {
    constructor(items, storage, fileSystem, options = {}) {
        this.items = [];
        this.metadata = {};
        this.initialMetadata = {};
        this.resumableDownloads = {};
        this.didSetItems = items ? true : false;
        this._setItems(items || []);
        this.storage = storage;
        this.storageKey = options.storageKey || exports.DEFAULT_STORAGE_KEY;
        this._downloadDirectory =
            options.downloadDirectory || exports.DEFAULT_DOWNLOAD_DIRECTORY;
        this._platform = options.platform;
        this.fileSystem = fileSystem;
        this.ready = false;
        this.didRestore = this.restore()
            .then(() => {
            return fileSystem.makeDirectoryAsync(`${fileSystem.documentDirectory}${this._downloadDirectory}`, {
                intermediates: true,
            });
        })
            .then(() => {
            this.ready = true;
            this.deleteUnusedFiles();
        });
    }
    hasItem(item) {
        return this.items.includes(item);
    }
    setItems(items) {
        this.didSetItems = true;
        this._setItems(items);
    }
    _setItems(items) {
        if (items.some((uri) => !(typeof uri === "string"))) {
            console.log("DownloadManager: Invalid uri in playlist", items);
            items = items.filter((uri) => typeof uri === "string");
        }
        if (items.length === this.items.length &&
            items.every((value, index) => value === this.items[index])) {
            return;
        }
        this.items = items;
        this._ensureMetadataForItems();
        // Remove unused metadata and resumable downloads
        Object.keys(this.metadata).forEach((uri) => {
            if (!this.items.includes(uri)) {
                delete this.resumableDownloads[uri];
                delete this.metadata[uri];
            }
        });
    }
    _ensureMetadataForItems() {
        this.items.forEach((item) => {
            if (!this.metadata[item]) {
                this.metadata[item] = {
                    downloadStatus: DOWNLOAD_STATUS.NOT_STARTED,
                };
            }
        });
    }
    getItemsWithDownloadStatus(downloadStatus) {
        return this.items.filter((item) => this.metadata[item] &&
            this.metadata[item].downloadStatus === downloadStatus);
    }
    setDownloadStatus(item, downloadStatus) {
        this.metadata[item].downloadStatus = downloadStatus;
    }
    getDownloadStatus(item) {
        return this.metadata[item].downloadStatus;
    }
    getMetadata(item) {
        return this.metadata[item];
    }
    setMetadata(item, metadata) {
        this.metadata[item] = metadata;
    }
    getResumableDownload(item) {
        return this.resumableDownloads[item];
    }
    setResumableDownload(item, resumableDownload) {
        if (resumableDownload) {
            this.resumableDownloads[item] = resumableDownload;
        }
        else {
            delete this.resumableDownloads[item];
        }
    }
    async restore() {
        if (!this.storageKey || this._platform === "web")
            return;
        try {
            const item = await this.storage.getItem(this.storageKey);
            if (item) {
                const metadata = JSON.parse(item);
                const documentPaths = await getDocumentPaths(this.fileSystem, this._downloadDirectory);
                Object.entries(metadata).forEach(([uri, itemMetadata]) => {
                    // Delete any items where we have a fileUri but the file doesn't
                    // exist on disk. This can happen if the user upgrades their app
                    // or if their document directory gets wiped for reasons outside
                    // of our control.
                    if (itemMetadata.fileUri &&
                        !documentPaths.includes(itemMetadata.fileUri)) {
                        delete metadata[uri];
                    }
                });
                this.metadata = metadata;
                this.initialMetadata = JSON.parse(item);
                this._ensureMetadataForItems();
                if (!this.didSetItems) {
                    this.items = Object.entries(metadata).map(([uri]) => uri);
                }
            }
        }
        catch (e) {
            console.log("DownloadManager: Failed to restore downloader metadata.");
        }
    }
    async save() {
        if (!this.storageKey || this._platform === "web")
            return;
        await this.didRestore;
        try {
            this.items.forEach((uri) => {
                const resumableDownload = this.resumableDownloads[uri];
                if (resumableDownload) {
                    this.metadata[uri].savedDownload = resumableDownload.savable();
                }
            });
            const json = JSON.stringify(this.metadata, null, 2);
            await this.storage.setItem(this.storageKey, json);
        }
        catch (e) {
            console.error("DownloadManager: Failed to save downloader metadata.", e);
        }
    }
    getCurrentFileUris() {
        return this.items.flatMap((uri) => {
            const itemMetadata = this.metadata[uri];
            return itemMetadata && itemMetadata.fileUri ? [itemMetadata.fileUri] : [];
        });
    }
    async deleteUnusedFiles() {
        if (this._platform === "web")
            return;
        const documentPaths = await getDocumentPaths(this.fileSystem, this._downloadDirectory);
        const currentFileUris = this.getCurrentFileUris();
        const filesToDelete = documentPaths
            .filter((path) => path.endsWith(".mp4"))
            .filter((path) => !currentFileUris.includes(path));
        const promises = filesToDelete.map(async (uri) => {
            try {
                return this.fileSystem.deleteAsync(uri, { idempotent: true });
            }
            catch (e) {
                console.log("DownloadManager: Failed to deleteAsync", uri, e);
            }
        });
        await promises;
    }
    _getStatus(downloadItems) {
        const remainingItemsCount = countFiles(downloadItems, DOWNLOAD_STATUS.NOT_STARTED) +
            countFiles(downloadItems, DOWNLOAD_STATUS.IN_PROGRESS);
        return {
            files: downloadItems,
            status: downloadItems.length === 0
                ? DOWNLOAD_STATUS.NOT_STARTED
                : countFiles(downloadItems, DOWNLOAD_STATUS.NOT_STARTED) ===
                    downloadItems.length
                    ? DOWNLOAD_STATUS.NOT_STARTED
                    : remainingItemsCount > 0
                        ? DOWNLOAD_STATUS.IN_PROGRESS
                        : DOWNLOAD_STATUS.COMPLETED,
            progress: downloadItems.length > 0
                ? (downloadItems.length - remainingItemsCount) / downloadItems.length
                : 0,
        };
    }
    /**
     * Get the status of the initial playlist restored from storage
     */
    getRestoredStatus() {
        if (this._platform === "web") {
            return { files: [], status: DOWNLOAD_STATUS.NOT_STARTED };
        }
        const downloadItems = Object.entries(this.initialMetadata).map(([uri, itemMetadata]) => createDownloadItem(uri, itemMetadata));
        return this._getStatus(downloadItems);
    }
    /**
     * Get the status of the current playlist
     */
    getStatus() {
        if (this._platform === "web") {
            return { files: [], status: DOWNLOAD_STATUS.NOT_STARTED };
        }
        const downloadItems = this.items.map((uri) => createDownloadItem(uri, this.getMetadata(uri)));
        return this._getStatus(downloadItems);
    }
}
exports.Playlist = Playlist;
